#!/usr/bin/python3

import sys

from lintian_brush.fixer import fixed_lintian_tag, report_result

EXPECTED_URL = (
    b"https://www.debian.org/doc/packaging-manuals/copyright-format/1.0/"
)

try:
    with open("debian/copyright", "rb") as f:
        lines = list(f.readlines())
    if not lines:
        sys.exit(2)
except FileNotFoundError:
    pass
else:
    import re

    m = re.fullmatch(b"(Format|Format-Specification): (.*)", lines[0].rstrip())
    if m and (
        m.group(1) != b"Format"
        or m.group(2).rstrip(b"/") != EXPECTED_URL.rstrip(b"/")
    ):
        fixed_lintian_tag(
            "source", "unversioned-copyright-format-uri", m.group(2).decode()
        )
        lines[0] = b"Format: %s\n" % EXPECTED_URL
        with open("debian/copyright", "wb") as f:
            f.writelines(lines)

report_result("Use versioned copyright format URI.")



#!/usr/bin/python3

# Find a list of tags that might qualify for inclusion in
# INFO_FIXERS in lintian_brush/lintian_overrides.py

import argparse

from lintian_brush.lintian_overrides import INFO_FIXERS
from lintian_brush.udd import connect_udd_mirror

parser = argparse.ArgumentParser()
args = parser.parse_args()

conn = connect_udd_mirror()
with conn.cursor() as cursor:
    cursor.execute(
        "SELECT package_type, package, package_version, information "
        "FROM lintian WHERE tag = 'mismatched-override'"
    )

    tag_count = {}
    for row in cursor:
        (_pkg_type, _pkg, _version, info) = row
        tag = info.split(" ")[0]
        tag_count.setdefault(tag, 0)
        tag_count[tag] += 1

tags_with_location_info = set()

with conn.cursor() as cursor:
    cursor.execute("SELECT tag FROM lintian WHERE information LIKE '%%[%%]'")
    for (tag,) in cursor:
        tags_with_location_info.add(tag)


for tag, count in sorted(tag_count.items(), reverse=True, key=lambda k: k[1]):
    if tag not in tags_with_location_info:
        # Looks like there's no location info in this tag's info
        continue
    if tag in INFO_FIXERS:
        # We already have a fixer
        continue
    print(f"{tag:50}  {count}")



#!/usr/bin/python3

import os
import sys

from lintian_brush.fixer import LintianIssue, report_result
from lintian_brush.line_editor import LineEditor

patches_path = "debian/patches"
if not os.path.isdir(patches_path):
    sys.exit(0)

with open(os.path.join(patches_path, "series")) as series:
    patches = series.read().splitlines()

if not patches:
    sys.exit(0)

for patch in patches:
    path = os.path.join(patches_path, patch)

    with LineEditor(path) as f:
        for lineno, line in f:
            try:
                (key, value) = line.split(":", 1)
            except ValueError:
                # Not a key-value pair.
                continue
            if key.strip() == "Origin":
                # If the Origin field is present and looks like an email address, use it as the author.
                issue = LintianIssue(
                    target="source",
                    info=f"[{path}]",
                    tag="dep3-format-patch-author-or-from-is-better",
                )
                if issue.should_fix():
                    if "@" in value:
                        value = value.strip()
                        f[lineno] = f"Author: {value}\n"
                    issue.report_fixed()
                break


report_result(
    "Use Author instead of Origin in patch headers.", certainty="confident"
)



#!/usr/bin/python3

import sys
from contextlib import suppress

from debmutate.watch import WatchEditor

from lintian_brush.fixer import LintianIssue, report_result

OBSOLETE_WATCH_FILE_FORMAT = 2
WATCH_FILE_LATEST_VERSION = 4


with suppress(FileNotFoundError), WatchEditor() as editor:
    if editor.watch_file is None:
        sys.exit(0)
    if editor.watch_file.version >= WATCH_FILE_LATEST_VERSION:
        pass
    else:
        if editor.watch_file.version <= OBSOLETE_WATCH_FILE_FORMAT:
            tag = "obsolete-debian-watch-file-standard"
        else:
            tag = "older-debian-watch-file-standard"
        issue = LintianIssue("source", tag, f"{editor.watch_file.version}")
        if issue.should_fix():
            editor.watch_file.version = WATCH_FILE_LATEST_VERSION
            issue.report_fixed()


report_result(
    f"Update watch file format version to {WATCH_FILE_LATEST_VERSION}."
)



#!/usr/bin/python3

import os
import sys
from typing import List

from debmutate._rules import (
    check_cdbs,
)
from debmutate.control import (
    ensure_exact_version,
    format_relations,
    is_relation_implied,
    iter_relations,
    parse_relations,
)
from debmutate.debhelper import (
    read_debhelper_compat_file,
)

from lintian_brush.debhelper import highest_stable_compat_level
from lintian_brush.fixer import control, fixed_lintian_tag, report_result

if not os.path.exists("debian/compat"):
    sys.exit(0)

# Package currently stores compat version in debian/compat..

debhelper_compat_version = read_debhelper_compat_file("debian/compat")

# debhelper >= 11 supports the magic debhelper-compat build-dependency.

# Exclude cdbs, since it only knows to get the debhelper compat version
# from debian/compat.

# debhelper-compat is only supported for stable compat levels
# https://bugs.debian.org/1026252

if (
    debhelper_compat_version < 11
    or check_cdbs()
    or debhelper_compat_version > highest_stable_compat_level()
):
    sys.exit(0)

# Upgrade to using debhelper-compat, drop debian/compat file.
os.unlink("debian/compat")

# Assume that the compat version is set in Build-Depends
with control as updater:
    insert_position = None
    changed_fields = []
    for field in [
        "Build-Depends",
        "Build-Depends-Indep",
        "Build-Depends-Arch",
    ]:
        to_delete: List[int] = []
        for offset, relation in iter_relations(
            updater.source.get(field, ""), "debhelper"
        ):
            if field == "Build-Depends" and {r.name for r in relation} == {
                "debhelper"
            }:
                # In the simple case, we'd just replace the debhelper
                # dependency with a debhelper-compat one, so remember the
                # location.
                insert_position = offset - len(to_delete)
            if is_relation_implied(
                relation, f"debhelper (>= {debhelper_compat_version})"
            ):
                to_delete.append(offset)

        if to_delete:
            # TODO(jelmer): Move this into a helper function in
            # lintian_brush.control.
            relations = parse_relations(updater.source[field])
            for i in reversed(to_delete):
                if i == 0 and len(relations) > 1:
                    # If the first item is removed, then copy the spacing to
                    # the next item
                    relations[1] = (
                        relations[0][0],
                        relations[1][1],
                        relations[0][2],
                    )
                del relations[i]

            updater.source[field] = format_relations(relations)
            changed_fields.append(field)

    updater.source["Build-Depends"] = ensure_exact_version(
        updater.source.get("Build-Depends", ""),
        "debhelper-compat",
        f"{debhelper_compat_version}",
        position=insert_position,
    )
    fixed_lintian_tag(updater.source, "uses-debhelper-compat-file", ())

    for field in changed_fields:
        if updater.source.get(field) == "":
            del updater.source[field]

report_result("Set debhelper-compat version in Build-Depends.")



#!/usr/bin/python3

from lintian_brush.fixer import control, fixed_lintian_tag, report_result

FIXABLE_HOSTS = [
    "gitlab.com",
    "github.com",
    "salsa.debian.org",
    "gitorious.org",
]


with control as updater:
    vcs_git = updater.source.get("Vcs-Git")
    if vcs_git and ":" in vcs_git:
        (netloc, path) = vcs_git.split(":", 1)
        if netloc.startswith("git@"):
            netloc = netloc[4:]
        if netloc in FIXABLE_HOSTS:
            fixed_lintian_tag(
                updater.source,
                "vcs-field-uses-not-recommended-uri-format",
                info="vcs-git {}".format(updater.source["Vcs-Git"]),
            )
            updater.source["Vcs-Git"] = f"https://{netloc}/{path}"


report_result("Use recommended URI format in Vcs header.")



#!/usr/bin/python3

import os

from debmutate.changelog import (
    WIDTH,
    ChangelogEditor,
    any_long_lines,
    rewrap_changes,
)

from lintian_brush.fixer import fixed_lintian_tag, report_result

updated = []


def wrap_block_lines(changes):
    if not any_long_lines(changes):
        return changes
    return rewrap_changes(changes)


def wrap_block(changelog, i):
    new_changes = wrap_block_lines(changelog[i].changes())
    if new_changes != changelog[i].changes():
        if i == 0:
            for lineno, change in enumerate(changelog[i].changes(), 2):
                if len(change) <= WIDTH:
                    continue
                # Lintian only warns about the first block.
                fixed_lintian_tag(
                    "source",
                    "debian-changelog-line-too-long",
                    info=f"line {lineno}"
                )
        changelog[i]._changes = new_changes
        updated.append(changelog[i].version)
        return True
    return False


with ChangelogEditor() as updater:
    if "CHANGELOG_THOROUGH" not in os.environ:
        wrap_block(updater.changelog, 0)
    else:
        for i in range(len(updater.changelog)):
            wrap_block(updater.changelog, i)


report_result(
    "Wrap long lines in changelog entries: {}.".format(
        ", ".join([str(v) for v in updated])
    )
)



#!/usr/bin/python3

from email.utils import parseaddr
from typing import Dict, List, Optional, Tuple

from lintian_brush.fixer import control, fixed_lintian_tag, report_result

REPLACEMENTS: Dict[Optional[str], Tuple[str, List[Tuple[str, str]]]] = {
    "python-modules-team@lists.alioth.debian.org": (
        "old-dpmt-vcs",
        [
            (
                "https://salsa.debian.org/python-team/modules/",
                "https://salsa.debian.org/python-team/packages/",
            )
        ],
    ),
    "python-apps-team@lists.alioth.debian.org": (
        "old-papt-vcs",
        [
            (
                "https://salsa.debian.org/python-team/applications/",
                "https://salsa.debian.org/python-team/packages/",
            )
        ],
    ),
}

with control as editor:
    maint: Optional[str]
    email: Optional[str]
    try:
        maint, email = parseaddr(editor.source["Maintainer"])
    except KeyError:
        maint, email = None, None
    changed_fields = set()
    try:
        tag, replacements = REPLACEMENTS[email]
    except KeyError:
        pass
    else:
        for field in [f for f in editor.source if f.startswith("Vcs-")]:
            url = editor.source[field]
            for old, new in replacements:
                url = url.replace(old, new)
            if url != editor.source[field]:
                editor.source[field] = url
                changed_fields.add(field)
                fixed_lintian_tag(editor.source, tag, info="")

    report_result(
        "Update fields {} for maintainer {}.".format(
            ", ".join(sorted(changed_fields)), maint
        )
    )



#!/usr/bin/python3

import re

from lintian_brush.fixer import report_result
from lintian_brush.upstream_metadata import ADDON_ONLY_FIELDS
from lintian_brush.yaml import YamlUpdater

SEP_CHARS = r"\n+|\s\s+|\t+"

obsolete_fields = {}
removed_fields = []

with YamlUpdater("debian/upstream/metadata") as editor:
    # If the debian/copyright file is machine-readable, then we can drop the
    # Name/Contact information from the debian/upstream/metadata file.
    if "Name" in editor.code or "Contact" in editor.code:
        from debmutate.copyright import upstream_fields_in_copyright

        obsolete_fields.update(
            upstream_fields_in_copyright("debian/copyright")
        )

    for field in ["Name", "Contact"]:
        try:
            um_value = editor.code[field]
        except KeyError:
            continue

        if um_value is None:
            del editor.code[field]
            removed_fields.append(field)
            continue

    for field, copyright_value in obsolete_fields.items():
        try:
            um_value = editor.code[field]
        except KeyError:
            continue
        if isinstance(copyright_value, tuple):
            copyright_entries = [x.strip() for x in copyright_value]
        else:
            copyright_entries = re.split(SEP_CHARS, copyright_value)
        um_entries = re.split(SEP_CHARS, um_value)
        um_entries = [x.lower() for x in um_entries]
        copyright_entries = [x.lower() for x in copyright_entries]
        if set(um_entries) == set(copyright_entries):
            del editor.code[field]
            removed_fields.append(field)

    if removed_fields and not (
        set(editor.code.keys()) - set(ADDON_ONLY_FIELDS)
    ):
        editor.code.clear()


report_result(
    "Remove obsolete field{} {} from debian/upstream/metadata "
    "(already present in machine-readable debian/copyright).".format(
        "s" if len(removed_fields) > 1 else "",
        ", ".join(sorted(removed_fields)),
    )
)



#!/usr/bin/python3

from lintian_brush.fixer import LintianIssue, control, report_result

binary_homepages = set()
source_homepage = None


with control as updater:
    source_homepage = updater.source.get("Homepage")
    for binary in updater.binaries:
        if "Homepage" not in binary:
            continue
        if source_homepage == binary["Homepage"]:
            issue = LintianIssue("source", "homepage-in-binary-package")
            # Source and binary both have a homepage field, but they're the
            # same ⇒ drop the binary package Homepage field
            if issue.should_fix():
                issue.report_fixed()
                del binary["Homepage"]
        else:
            binary_homepages.add(binary["Homepage"])

    if (
        source_homepage is None
        and binary_homepages
        and len(binary_homepages) == 1
    ):
        updater.source["Homepage"] = binary_homepages.pop()

        for binary in updater.binaries:
            if "Homepage" in binary:
                issue = LintianIssue("source", "homepage-in-binary-package")
                if issue.should_fix():
                    del binary["Homepage"]
                    issue.report_fixed()


report_result("Set Homepage field in Source rather than Binary package.")



#!/usr/bin/python3

from lintian_brush.fixer import LintianIssue, control, report_result

updated_packages = set()


with control as updater:
    for binary in updater.binaries:
        package = binary["Package"]
        if not package.startswith("fonts-") and not package.startswith(
            "xfonts-"
        ):
            continue
        if binary.get("Architecture") not in ("all", None):
            continue
        if "Multi-Arch" in binary:
            continue
        issue = LintianIssue(
            updater.source, "font-package-not-multi-arch-foreign"
        )
        if issue.should_fix():
            binary["Multi-Arch"] = "foreign"
            updated_packages.add(package)
            issue.report_fixed()


report_result(
    "Set Multi-Arch: foreign on package{} {}.".format(
        "s" if len(updated_packages) > 1 else "", ", ".join(updated_packages)
    )
)



#!/usr/bin/python3

from email.utils import parseaddr

from lintian_brush.fixer import LintianIssue, control, report_result

with control as editor:
    try:
        old_maintainer = editor.source["Maintainer"]
    except KeyError:
        pass
    else:
        maint, email = parseaddr(old_maintainer)
        if email in (
            "python-modules-team@lists.alioth.debian.org",
            "python-modules-team@alioth-lists.debian.net",
            "python-apps-team@lists.alioth.debian.org",
        ):
            issue = LintianIssue(
                editor.source, "python-teams-merged", info=(old_maintainer,)
            )
            if issue.should_fix():
                editor.source["Maintainer"] = (
                    "Debian Python Team <team+python@tracker.debian.org>"
                )
                issue.report_fixed()

report_result("Update maintainer email for merge of DPMT and PAPT.")



#!/usr/bin/python3

import os

from lintian_brush.fixer import LintianIssue, report_result
from lintian_brush.xdg import DesktopEntryEditor

paths = []

for name in os.listdir("debian"):
    if name.endswith(".desktop"):
        path = os.path.join("debian", name)
        with DesktopEntryEditor(path) as updater:
            if updater.get("Encoding") == "UTF-8":
                issue = LintianIssue(
                    "source",
                    "desktop-entry-contains-encoding-key",
                    "XX:YY Encoding",
                )
                if issue.should_fix():
                    del updater.entry["Encoding"]
                    paths.append(path)
                    issue.report_fixed()
            # TODO(jelmer): if encoding is non-UTF-8, invoke iconv.


if len(paths) == 1:
    report_result(
        f"Remove deprecated Encoding key from desktop file {paths[0]}."
    )
else:
    report_result(
        "Remove deprecated Encoding key from desktop files: {}.".format(
            ", ".join(sorted(paths))
        )
    )



#!/usr/bin/python3

from lintian_brush.fixer import fixed_lintian_tag, report_result
from lintian_brush.systemd import SystemdServiceEditor, systemd_service_files

for path in systemd_service_files():
    with SystemdServiceEditor(path) as updater:
        if (
            updater.file["Unit"]["DefaultDependencies"] == "no"
            and "shutdown.target" in updater.file["Unit"]["Conflicts"]
            and "shutdown.target" not in updater.file["Unit"]["Before"]
        ):
            updater.file["Unit"]["Before"].append("shutdown.target")
            fixed_lintian_tag(
                "source", "systemd-service-file-shutdown-problems", path
            )

report_result("Add Before=shutdown.target to Unit section.")



#!/usr/bin/python3
# Copyright (C) 2019 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with YAML files."""

import json
import os
from io import StringIO
from typing import Any, Iterator, List

from ruamel.yaml import YAML


class MultiYamlUpdater:
    def __init__(self, path: str, remove_empty: bool = False):
        self.yaml = YAML()
        self.path = path
        self.remove_empty = remove_empty
        self._preamble: List[str] = []
        self._dirpath = os.path.dirname(path)
        self._force_rewrite = False

    def __getitem__(self, i):
        return self._code[i]

    def __setitem__(self, i, val):
        self._code[i] = val

    def __iter__(self) -> Iterator[Any]:
        yield from self._code

    def __delitem__(self, i):
        del self._code[i]

    def __len__(self):
        return len(self._code)

    def __enter__(self) -> "MultiYamlUpdater":
        try:
            with open(self.path) as f:
                inp = list(f)
        except FileNotFoundError:
            self._orig: List[Any] = []
            self._orig_text = ""
        else:
            for line in inp:
                if line != "\n" and not line.startswith("#"):
                    break
                self._preamble.append(line)
            self._orig_text = "".join(inp[len(self._preamble) :])
            self._orig = list(self.yaml.load_all(self._orig_text))
        self._code = self._orig.copy()
        return self

    def _get_code(self):
        return self._code

    def _set_code(self, v):
        self._code = v

    code = property(_get_code, _set_code)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not exc_type:
            if not self._code and self.remove_empty:
                if os.path.exists(self.path):
                    os.unlink(self.path)
                    if self._dirpath and not os.listdir(self._dirpath):
                        os.rmdir(self._dirpath)
            else:
                if self._force_rewrite or self._code != self._orig:
                    if not os.path.exists(self._dirpath) and self._dirpath:
                        os.mkdir(self._dirpath)
                    with open(self.path, "w") as f:
                        f.writelines(self._preamble)
                        self.yaml.dump_all(self._code, f)
        return False


class YamlUpdater:
    def __init__(
        self,
        path: str,
        remove_empty: bool = True,
        allow_duplicate_keys: bool = False,
    ):
        self.yaml = YAML()
        self.yaml.allow_duplicate_keys = allow_duplicate_keys
        self.path = path
        self._dirpath = os.path.dirname(path)
        self.remove_empty = remove_empty
        self._directives: List[str] = []
        self._preamble: List[str] = []
        self._force_rewrite = False

    def force_rewrite(self):
        self._force_rewrite = True

    def __enter__(self):
        try:
            with open(self.path) as f:
                inp = list(f)
        except FileNotFoundError:
            self._orig = {}
            self._orig_text = ""
        else:
            for line in inp:
                if line != "\n" and not line.startswith("#"):
                    break
                self._preamble.append(line)
            if "---\n" in inp:
                for i, line in enumerate(inp):
                    if line == "---\n":
                        self._directives = inp[len(self._preamble) : i + 1]
                        break
            self._orig_text = "".join(inp[len(self._preamble) :])
            self._orig = self.yaml.load(self._orig_text)
        self._code = self._orig.copy()
        return self

    def _get_code(self):
        return self._code

    def _set_code(self, v):
        self._code = v

    code = property(_get_code, _set_code)

    def _only_simple_changes(self):
        if not isinstance(self._orig, dict):
            return False

        def is_one_line(k, v):
            f = StringIO()
            self.yaml.dump({k: v}, f)
            return len(f.getvalue().splitlines()) == 1

        # Check if there are only "simple" changes, i.e.
        # changes that update string fields
        for k, v in self._code.items():
            if v == self._orig.get(k) or not is_one_line(k, v):
                continue
        for k, v in self._orig.items():
            if k not in self._code and not is_one_line(k, v):
                return False
            if ":" in k:
                return False
        return True

    def _update_lines(self, lines, f):
        for line in self._directives:
            f.write(line)
        if "".join(lines[len(self._directives) :]).startswith("{"):
            _update_json_lines(
                self._orig, self._code, lines[len(self._directives) :], f
            )
        else:
            _update_yaml_lines(
                self.yaml,
                self._orig,
                self._code,
                lines[len(self._directives) :],
                f,
            )

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not exc_type:
            if not self._code and self.remove_empty:
                if os.path.exists(self.path):
                    os.unlink(self.path)
                    if self._dirpath and not os.listdir(self._dirpath):
                        os.rmdir(self._dirpath)
            else:
                if self._force_rewrite or self._code != self._orig:
                    if not os.path.exists(self._dirpath) and self._dirpath:
                        os.mkdir(self._dirpath)
                    if not self._force_rewrite and self._only_simple_changes():
                        try:
                            with open(self.path) as f:
                                lines = list(f.readlines())
                        except FileNotFoundError:
                            lines = ["---\n"]
                        with open(self.path, "w") as f:
                            self._update_lines(lines, f)
                    else:
                        with open(self.path, "w") as f:
                            f.writelines(self._preamble)
                            if self._force_rewrite:
                                f.writelines(self._directives)
                            self.yaml.dump(self._code, f)
        return False


def update_ordered_dict(code, changed, key=None):
    if key is None:

        def key(x):
            return x

    to_insert = []
    for k, v in changed:
        if k in code:
            code[k] = v
        else:
            to_insert.append((k, v))

    to_insert.sort(key=key)
    i = 0  # noqa: SIM113
    for k, v in list(code.items()):
        while to_insert and key((k, v)) > key(to_insert[0]):
            code.insert(i, *to_insert.pop(0))
            i += 1
        i += 1

    while to_insert:
        k, v = to_insert.pop(0)
        code[k] = v


def _update_json_lines(orig, code, lines, f):
    indent = 0  # noqa: SIM113
    for c in lines[1][0:]:
        if c != " ":
            break
        indent += 1
    json.dump(code, f, indent=indent)
    f.write("\n")


def _update_yaml_lines(yaml, orig, code, lines, f):
    os = list(orig.keys())
    cs = list(code.keys())
    o = 0
    for line in lines:
        try:
            key = os[o]
        except IndexError:
            key = None
        if key and ":" in line and line.split(":", 1)[0].strip() == key:
            while cs and cs[0] not in os:
                yaml.dump({cs[0]: code[cs[0]]}, f)
                cs.pop(0)
            if key not in code:
                pass  # Line was removed
            elif code[key] == orig[key]:
                # Line stayed the same
                f.write(line)
                cs.remove(key)
            else:
                yaml.dump({key: code[key]}, f)
                cs.remove(key)
            o += 1
        else:
            f.write(line)
    if cs and not line.endswith("\n"):
        f.write("\n")
    while cs:
        key = cs.pop(0)
        yaml.dump({key: code[key]}, f)



#!/usr/bin/python3

import os
import sys

from lintian_brush.fixer import report_result
from lintian_brush.gpg import GpgMissing, gpg_import_export

try:
    with open("debian/upstream/signing-key.pgp", "rb") as f:
        key = f.read()
except FileNotFoundError:
    sys.exit(0)

try:
    full = gpg_import_export(
        [
            "no-import-minimal",
            "no-import-clean",
            "no-self-sigs-only",
            "no-repair-keys",
            "import-restore",
        ],
        [],
        key,
    )
except GpgMissing:
    sys.exit(2)

with open("debian/upstream/signing-key.asc", "wb") as f:
    f.write(full)

os.remove("debian/upstream/signing-key.pgp")

report_result("Enarmor upstream signing key.")



#!/usr/bin/python3

import os
import sys

from lintian_brush.fixer import (
    LintianIssue,
    control,
    report_result,
)
from lintian_brush.lintian import LINTIAN_DATA_PATH


def parse(t):
    return tuple([int(v) for v in t.split(".")])


PYTHON_VERSIONS_PATH = os.path.join(LINTIAN_DATA_PATH, "python/versions")


python_versions = {}
try:
    with open(os.path.join(PYTHON_VERSIONS_PATH)) as f:
        for line in f:
            line = line.strip()
            if not line:
                continue
            key, value = line.split("=")
            python_versions[key.strip()] = parse(value.strip())
except FileNotFoundError:
    sys.exit(2)


with control as updater:
    # Remove anything that involves python 2.6, 2.7, 3.3
    if "X-Python-Version" in updater.source and updater.source[
        "X-Python-Version"
    ].strip().startswith(">= 2."):
        vers = updater.source["X-Python-Version"].split(">=")[1].strip()
        if parse(vers) <= python_versions["old-python2"]:
            if parse(vers) <= python_versions["ancient-python2"]:
                kind = "ancient"
            else:
                kind = "old"
        issue = LintianIssue(
            updater.source,
            f"{kind}-python-version-field",
            "x-python-version {}".format(updater.source["X-Python-Version"]),
        )
        if issue.should_fix():
            del updater.source["X-Python-Version"]
            issue.report_fixed()
    if "X-Python3-Version" in updater.source and updater.source[
        "X-Python3-Version"
    ].strip().startswith(">="):
        vers = updater.source["X-Python3-Version"].split(">=")[1].strip()
        if parse(vers) <= python_versions["old-python3"]:
            if parse(vers) <= python_versions["ancient-python3"]:
                kind = "ancient"
            else:
                kind = "old"
            issue = LintianIssue(
                updater.source,
                f"{kind}-python-version-field",
                "x-python3-version {}".format(
                    updater.source["X-Python3-Version"]
                ),
            )
            if issue.should_fix():
                del updater.source["X-Python3-Version"]
                issue.report_fixed()


report_result(
    "Remove unnecessary X-Python{,3}-Version field in debian/control."
)



#!/usr/bin/python3

import os
import sys

from debmutate.control import ensure_some_version
from debmutate.debhelper import get_sequences

from lintian_brush.fixer import (
    LintianIssue,
    control,
    meets_minimum_certainty,
    report_result,
)

CERTAINTY = "possible"


if not meets_minimum_certainty(CERTAINTY):
    sys.exit(2)


with control:
    sequences = list(get_sequences(control_editor=control))

    if "nodejs" not in sequences:
        sys.exit(2)

    issue = LintianIssue(
        control.source,
        "pkg-js-tools-test-is-missing",
        "debian/tests/pkg-js/test",
    )
    if issue.should_fix():
        if os.path.exists("test/node.js"):
            os.makedirs("debian/tests", exist_ok=True)
            with open("debian/tests/pkg-js/test", "w") as f:
                f.write("mocha test/node.js\n")
            control.source["Build-Depends"] = ensure_some_version(
                control.source.get("Build-Depends", ""), "mocha <!nocheck>"
            )
            issue.report_fixed()
        elif os.path.exists("test.js"):
            os.makedirs("debian/tests/pkg-js", exist_ok=True)
            with open("debian/tests/pkg-js/test", "w") as f:
                f.write("tape test.js\n")
            control.source["Build-Depends"] = ensure_some_version(
                control.source.get("Build-Depends", ""), "node-tape <!nocheck>"
            )
            issue.report_fixed()


report_result("Add autopkgtest for node.", certainty=CERTAINTY)



#!/usr/bin/python3

import os
import sys

from debmutate._rules import RulesEditor

from lintian_brush.fixer import LintianIssue, control, report_result


def get_archs():
    archs = set()

    # TODO(jelmer): Not really an 'update'
    with control as updater:
        for binary in updater.binaries:
            archs.add(binary["Architecture"])
    return archs


added = []


if not os.path.exists("debian/rules"):
    sys.exit(2)
with RulesEditor() as editor:
    has_build_arch = bool(
        list(editor.makefile.iter_rules(b"build-arch", exact=False))
    )
    has_build_indep = bool(
        list(editor.makefile.iter_rules(b"build-indep", exact=False))
    )

    if has_build_arch and has_build_indep:
        sys.exit(0)

    if any(
        line.lstrip(b" -").startswith(b"include ")
        for line in editor.makefile.dump_lines()
    ):
        # No handling of includes for the moment.
        sys.exit(0)

    archs = get_archs()
    if not has_build_indep:
        issue = LintianIssue(
            "source",
            "debian-rules-missing-recommended-target",
            info="build-indep",
        )
        if issue.should_fix():
            added.append("build-indep")
            editor.makefile.add_rule(
                b"build-indep",
                components=([b"build"] if "all" in archs else None),
            )
            issue.report_fixed()
    if not has_build_arch:
        added.append("build-arch")
        issue = LintianIssue(
            "source",
            "debian-rules-missing-recommended-target",
            info="build-arch",
        )
        if issue.should_fix():
            editor.makefile.add_rule(
                b"build-arch",
                components=([b"build"] if (archs - {"all"}) else None),
            )
            issue.report_fixed()

    if not added:
        sys.exit(0)

    try:
        phony_rule = list(editor.makefile.iter_rules(b".PHONY"))[-1]
    except IndexError:
        pass
    else:
        for c in added:
            phony_rule.append_component(c.encode())


if len(added) == 1:
    report_result(f"Add missing debian/rules target {added[0]}.")
else:
    report_result(
        "Add missing debian/rules targets {}.".format(", ".join(added))
    )



#!/usr/bin/python3

from debmutate.control import (
    add_dependency,
    drop_dependency,
    get_relation,
)

from lintian_brush.fixer import LintianIssue, control, report_result

with control as updater:
    try:
        get_relation(
            updater.source.get("Build-Depends-Indep", ""),
            "libmodule-build-perl",
        )
    except KeyError:
        pass
    else:
        issue = LintianIssue(
            updater.source,
            "libmodule-build-perl-needs-to-be-in-build-depends",
            "",
        )
        if issue.should_fix():
            updater.source["Build-Depends-Indep"] = drop_dependency(
                updater.source["Build-Depends-Indep"], "libmodule-build-perl"
            )
            updater.source["Build-Depends"] = add_dependency(
                updater.source.get("Build-Depends", ""), "libmodule-build-perl"
            )
            issue.report_fixed()


report_result(
    "Move libmodule-build-perl from Build-Depends-Indep to Build-Depends."
)



#!/usr/bin/python3

from typing import Optional
from urllib.parse import urlparse

from upstream_ontologist.guess import (
    guess_upstream_metadata_items,
    known_bad_guess,
)

from lintian_brush.fixer import (
    LintianIssue,
    control,
    meets_minimum_certainty,
    report_result,
    trust_package,
)


def guess_homepage():
    for datum in guess_upstream_metadata_items(
        ".", trust_package=trust_package()
    ):
        if datum.field != "Homepage":
            continue
        if not meets_minimum_certainty(datum.certainty):
            continue
        if known_bad_guess(datum):
            continue
        return datum


with control as updater:
    issue: Optional[LintianIssue]
    if "Homepage" not in updater.source:
        datum = guess_homepage()
        issue = LintianIssue("source", "no-homepage-field")
        if datum and issue.should_fix():
            updater.source["Homepage"] = datum.value
            issue.report_fixed()
            report_result("Fill in Homepage field.", certainty=datum.certainty)
    else:
        hostname = urlparse(updater.source["Homepage"]).hostname
        if hostname == "pypi.org":
            issue = LintianIssue(
                "source", "pypi-homepage", updater.source["Homepage"]
            )
        elif hostname == "rubygems.org":
            issue = LintianIssue(
                "source", "rubygem-homepage", updater.source["Homepage"]
            )
        else:
            issue = None

        if issue:
            datum = guess_homepage()
            if issue.should_fix() and datum:
                updater.source["Homepage"] = datum.value
                issue.report_fixed()
                report_result(
                    f"Avoid {hostname} in Homepage field.",
                    certainty=datum.certainty,
                )



#!/usr/bin/python3

from lintian_brush.fixer import fixed_lintian_tag, report_result
from lintian_brush.systemd import SystemdServiceEditor, systemd_service_files

for path in systemd_service_files():
    with SystemdServiceEditor(path) as updater:
        unit = updater.file["Unit"]
        try:
            unit.rename("BindTo", "BindsTo")
        except KeyError:
            pass
        else:
            fixed_lintian_tag(
                "source",
                "systemd-service-file-refers-to-obsolete-bindto",
                path,
            )


report_result("Rename BindTo key to BindsTo in systemd files.")



#!/usr/bin/python3

from lintian_brush.fixer import fixed_lintian_tag, report_result
from lintian_brush.systemd import SystemdServiceEditor, systemd_service_files

DEPRECATED_TARGETS = ["syslog.target"]

removed = []

for path in systemd_service_files():
    with SystemdServiceEditor(path) as updater:
        for target in DEPRECATED_TARGETS:
            try:
                updater.file["Unit"]["After"].remove(target)
            except ValueError:
                pass
            else:
                removed.append((path, target))
                fixed_lintian_tag(
                    "source",
                    "systemd-service-file-refers-to-obsolete-target",
                    (path, target),
                )

removed.sort()

report_result(
    "Remove references to obsolete targets in systemd unit files: {}.".format(
        ", ".join([f"{filename} ({target})" for (filename, target) in removed])
    )
)



#!/usr/bin/python3

import sys

from lintian_brush.fixer import LintianIssue, control, report_result

changed = []


try:
    with control as updater:
        default_priority = updater.source.get("Priority")

        for binary in updater.binaries:
            if binary.get("Section") != "libs":
                continue
            priority = binary.get("Priority", default_priority)
            if priority in ("required", "important", "standard"):
                issue = LintianIssue(
                    binary,
                    "excessive-priority-for-library-package",
                    info=priority,
                )
                if issue.should_fix():
                    binary["Priority"] = "optional"
                    changed.append(binary["Package"])
                    issue.report_fixed()
except FileNotFoundError:
    sys.exit(0)


report_result(
    "Set priority for library package{} {} to optional.".format(
        "s" if len(changed) > 1 else "", ", ".join(changed)
    )
)



#!/usr/bin/python3

from debmutate._rules import RulesEditor
from debmutate.control import drop_dependency, get_relation

from lintian_brush.debhelper import add_sequence
from lintian_brush.fixer import LintianIssue, control, report_result

with control:
    for binary in control.binaries:
        try:
            rel = get_relation(binary.get("Depends", ""), "vim-addon-manager")
        except KeyError:
            continue
        issue = LintianIssue(binary, "obsolete-vim-addon-manager")
        if not issue.should_fix():
            continue
        binary["Depends"] = drop_dependency(
            binary["Depends"], "vim-addon-manager"
        )
        with RulesEditor() as rules:
            add_sequence(control, rules, "vim-addon")
        issue.report_fixed()


report_result("Migrate from vim-addon-manager to dh-vim-addon.")



#!/usr/bin/python3

import os

from lintian_brush.fixer import fixed_lintian_tag, report_result

removed = []

for name in os.listdir("debian"):
    if name.endswith(".linda-overrides"):
        os.unlink(os.path.join("debian", name))
        removed.append(name)
        fixed_lintian_tag(
            "source",
            "package-contains-linda-override",
            "usr/share/linda/overrides/{}".format(
                name[: -len(".linda-overrides")]
            ),
        )

report_result("Remove obsolete linda overrides: " + ", ".join(removed))



#!/usr/bin/python3

import sys
from contextlib import suppress

from lintian_brush.fixer import report_result, warn
from lintian_brush.yaml import YamlUpdater

try:
    from Levenshtein import distance
except ImportError:
    sys.exit(2)

valid_field_names = {
    "Archive",
    "ASCL-Id",
    "Bug-Database",
    "Bug-Submit",
    "Cite-As",
    "Changelog",
    "CPE",
    "Documentation",
    "Donation",
    "FAQ",
    "Funding",
    "Gallery",
    "Other-References",
    "Reference",
    "Registration",
    "Registry",
    "Repository",
    "Repository-Browse",
    "Screenshots",
    "Security-Contact",
    "Webservice",
}

typo_fixed = set()
case_fixed = set()

with suppress(FileNotFoundError), YamlUpdater(
    "debian/upstream/metadata"
) as updater:
    for field in updater.code:
        if field in valid_field_names:
            continue
        if field.startswith("X-") and field[2:] in valid_field_names:
            if field[2:] in updater.code:
                warn(f"Both {field} and {field[2:]} exist.")
                continue
            value = updater.code[field]
            del updater.code[field]
            updater.code[field[2:]] = value
            typo_fixed.add((field, field[2:]))
            continue

        for option in valid_field_names:
            if distance(field, option) == 1:
                value = updater.code[field]
                del updater.code[field]
                updater.code[option] = value
                if option.lower() == field.lower():
                    case_fixed.add((field, option))
                else:
                    typo_fixed.add((field, option))
                break


if case_fixed:
    kind = "case" + ("s" if len(case_fixed) > 1 else "")
else:
    kind = ""
if typo_fixed:
    if case_fixed:
        kind += " and "
    kind += "typo" + ("s" if len(typo_fixed) > 1 else "")

fixed_str = ", ".join(
    [
        f"{old} ⇒ {new}"
        for (old, new) in sorted(list(case_fixed) + list(typo_fixed))
    ]
)

report_result(
    f"Fix field name {kind} in debian/upstream/metadata ({fixed_str})."
)



#!/usr/bin/python3

import os
import sys

from lintian_brush.fixer import (
    LintianIssue,
    opinionated,
    package_is_native,
    report_result,
)

if not package_is_native():
    # Nothing to do
    sys.exit(0)

if not opinionated():
    sys.exit(0)

for name in ["debian/upstream/signing-key.asc"]:
    if not os.path.exists(name):
        continue
    issue = LintianIssue("source", "public-upstream-key-in-native-package", ())
    if not issue.should_fix():
        continue
    issue.report_fixed()
    os.unlink(name)

try:
    if os.listdir("debian/upstream") == []:
        os.rmdir("debian/upstream")
except FileNotFoundError:
    pass

report_result(
    "Remove upstream signing key in native source package.",
    certainty="certain",
)



#!/usr/bin/python3

import os
import sys
from typing import List, Tuple

from debmutate._rules import RulesEditor

from lintian_brush.fixer import LintianIssue, report_result
from lintian_brush.lintian import dh_commands

if not os.path.exists("debian/rules"):
    sys.exit(2)

try:
    from Levenshtein import distance
except ModuleNotFoundError:
    sys.exit(2)

known_dh_commands = list(dh_commands())

# Include javahelper binaries, since some are just one character away from
# debhelper ones.
known_dh_commands.extend(
    [
        "jh_build",
        "jh_classpath",
        "jh_clean",
        "jh_compilefeatures",
        "jh_depends",
        "jh_exec",
        "jh_generateorbitdir",
        "jh_installeclipse",
        "jh_installjavadoc",
        "jh_installlibs",
        "jh_linkjars",
        "jh_makepkg",
        "jh_manifest",
        "jh_repack",
        "jh_setupenvironment",
        "mh_checkrepo",
        "mh_install",
        "mh_installpoms",
        "mh_linkjars",
        "mh_patchpoms",
        "mh_clean",
        "mh_installjar",
        "mh_installsite",
        "mh_linkrepojar",
        "mh_unpatchpoms",
        "mh_cleanpom",
        "mh_installpom",
        "mh_linkjar",
        "mh_patchpom",
    ]
)

known_targets = set()
for dh_command in known_dh_commands:
    known_targets.update(
        [
            "override_" + dh_command,
            "execute_before_" + dh_command,
            "execute_after_" + dh_command,
        ]
    )


renamed: List[Tuple[str, str]] = []

with RulesEditor() as editor:
    for rule in editor.makefile.iter_all_rules():
        if rule.target.decode() in known_targets:
            continue
        for known_target in known_targets:
            issue = LintianIssue(
                "source",
                "typo-in-debhelper-override-target",
                "%s -> %s (line X)",
            )
            if (
                distance(known_target, rule.target.decode()) == 1
                and issue.should_fix()
            ):
                renamed.append((rule.target.decode(), known_target))
                rule.rename_target(rule.target, known_target.encode())
                issue.report_fixed()


report_result(
    "Fix typo in debian/rules rules: {}".format(
        ", ".join(f"{old} ⇒ {new}" for old, new in renamed)
    )
)



#!/usr/bin/python3

import os
import sys
from typing import Optional

from debmutate.control import ensure_some_version

from lintian_brush.fixer import (
    control,
    fixed_lintian_tag,
    is_debcargo_package,
    report_result,
)

if is_debcargo_package():
    sys.exit(0)

format: Optional[str]
try:
    with open("debian/source/format") as f:
        format = f.read().strip()
except FileNotFoundError:
    format = None

if format != "3.0 (quilt)" and os.path.exists("debian/patches/series"):
    with control as updater:
        updater.source["Build-Depends"] = ensure_some_version(
            updater.source.get("Build-Depends", ""), "quilt"
        )
    if updater.changed:
        fixed_lintian_tag(updater.source, "quilt-series-but-no-build-dep")


report_result("Add missing dependency on quilt.")



#!/usr/bin/python3

from email.utils import parseaddr

from lintian_brush.fixer import LintianIssue, control, report_result

with control as updater:
    if (
        "Maintainer" in updater.source
        and parseaddr(updater.source["Maintainer"])[1]
        == "packages@qa.debian.org"
        and "Uploaders" in updater.source
    ):
        issue = LintianIssue(updater.source, "uploaders-in-orphan", info=())
        if issue.should_fix():
            del updater.source["Uploaders"]
            issue.report_fixed()


report_result("Remove uploaders from orphaned package.")



#!/usr/bin/python3

import sys
from typing import List, Optional

from upstream_ontologist import UpstreamDatum
from upstream_ontologist.guess import guess_upstream_metadata

from lintian_brush.fixer import (
    LintianIssue,
    control,
    meets_minimum_certainty,
    net_access_allowed,
    report_result,
    trust_package,
)

CERTAINTY = "possible"

if not meets_minimum_certainty(CERTAINTY):
    sys.exit(0)


def textwrap_description(text: UpstreamDatum | str) -> List[str]:
    import textwrap

    if isinstance(text, UpstreamDatum):
        text = text.value

    ret = []
    paras = text.split("\n\n")
    for para in paras:
        if "\n*" in para:
            ret.extend(para.splitlines())
        else:
            ret.extend(textwrap.wrap(para))
        ret.append("")
    if not ret[-1]:
        ret.pop(-1)
    return ret


# TODO(jelmer): Use debmutate.control.format_description instead
def format_description(summary: UpstreamDatum | str, lines):
    if isinstance(summary, UpstreamDatum):
        summary = summary.value
    return summary + "\n" + "".join([f" {line}\n" for line in lines])


def guess_description(binary_name, all_binaries, summary=None):
    if len(all_binaries) != 1:
        # TODO(jelmer): Support handling multiple binaries
        return None

    upstream_metadata = guess_upstream_metadata(
        ".", trust_package(), net_access_allowed()
    )
    if summary is None:
        summary = upstream_metadata.get("Summary")
    try:
        upstream_description = textwrap_description(
            upstream_metadata["Description"]
        )
    except KeyError:
        # Better than nothing..
        return summary

    if summary is None:
        if len(upstream_description) == 1:
            return upstream_description[0].rstrip("\n")
        return None
    lines = [line if line else "." for line in upstream_description]

    return format_description(summary, lines)


updated = []

with control as updater:
    summary: Optional[str]
    for binary in updater.binaries:
        existing_description = binary.get("Description")
        if not existing_description:
            issue = LintianIssue(binary, "required-field", "Description")
            summary = None
        elif len(existing_description.strip().splitlines()) == 1:
            issue = LintianIssue(binary, "extended-description-is-empty")
            summary = existing_description.splitlines()[0]
        else:
            continue
        if not issue.should_fix():
            continue
        description = guess_description(
            binary["Package"], list(updater.binaries), summary=summary
        )
        if description and description != existing_description:
            binary["Description"] = description
            updated.append(binary["Package"])
            issue.report_fixed()


report_result(
    description="Add description for binary packages: {}".format(
        ", ".join(sorted(updated))
    ),
    certainty=CERTAINTY,
)



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import glob

from setuptools import setup
from setuptools_rust import Binding, RustBin, RustExtension

setup(
    data_files=[
        (
            "share/lintian-brush/fixers",
            [n for n in glob.glob("fixers/*") if not n.endswith("/slow")],
        ),
        (
            "share/lintian-brush",
            [
                "spdx.json",
                "lintian-brush/renamed-tags.json",
                "analyzer/key-package-versions.json",
            ],
        ),
    ],
    rust_extensions=[
        RustBin("debianize", "debianize/Cargo.toml"),
        RustBin("lintian-brush", "lintian-brush/Cargo.toml"),
        RustBin(
            "detect-changelog-behaviour",
            "analyzer/Cargo.toml",
            features=["cli"],
        ),
        RustBin("deb-vcs-publish", "analyzer/Cargo.toml", features=["cli"]),
        RustBin("dump-multiarch-hints", "multiarch-hints/Cargo.toml"),
        RustBin(
            "apply-multiarch-hints",
            "multiarch-hints/Cargo.toml",
            features=["cli"],
        ),
        RustExtension(
            "lintian_brush._lintian_brush_rs",
            "lintian-brush-py/Cargo.toml",
            binding=Binding.PyO3,
            features=["extension-module"],
        ),
        RustBin(
            "deb-scrub-obsolete", "scrub-obsolete/Cargo.toml", features=["cli"]
        ),
        RustBin(
            "deb-transition-apply",
            "transition-apply/Cargo.toml",
            features=["cli"],
        ),
    ],
)



#!/usr/bin/python3
# Report lintian tags that might be good candidates to implement fixers for.

import argparse

import psycopg2
from ruamel.yaml import YAML

from lintian_brush import fixable_lintian_tags

parser = argparse.ArgumentParser()
parser.add_argument(
    "--exclude",
    type=str,
    default="hard",
    help="Comma-separated list of difficulties to exclude.",
)
args = parser.parse_args()

conn = psycopg2.connect(
    "postgresql://udd-mirror:udd-mirror@udd-mirror.debian.net/udd"
)
with conn.cursor() as cursor:
    cursor.execute(
        "SELECT tag, COUNT(DISTINCT package) AS package_count, "
        "COUNT(*) AS tag_count from lintian "
        "WHERE tag_type NOT IN ('classification') GROUP BY 1 ORDER BY 2 DESC"
    )

    supported_tags = set(fixable_lintian_tags())

    yaml = YAML()
    with open("tag-status.yaml") as f:
        tag_status = yaml.load(f)

    per_tag_status = {}
    for entry in tag_status or []:
        per_tag_status[entry["tag"]] = entry

    exclude_difficulties = args.exclude.split(",")

    for tag, package_count, tag_count in cursor:
        if tag in supported_tags:
            continue
        difficulty = per_tag_status.get(tag, {}).get("difficulty", "unknown")
        if difficulty in exclude_difficulties:
            continue
        print(f"{tag} {difficulty} {package_count}/{tag_count}")



#!/usr/bin/python3

import sys

try:
    import psycopg2  # noqa: F401
except ModuleNotFoundError:
    sys.exit(2)

from debmutate.vcs import (
    split_vcs_url,
    unsplit_vcs_url,
)

from lintian_brush.fixer import (
    control,
    diligence,
    opinionated,
    report_result,
    warn,
)
from lintian_brush.vcs import determine_browser_url

if diligence() < 1:
    # Removing unused overrides requires pro-actively contacting UDD.
    sys.exit(0)

from lintian_brush.vcswatch import VcsWatch, VcsWatchError


def get_default_branch(url, branch=None):
    from dulwich.client import get_transport_and_path

    c, p = get_transport_and_path(url)
    result = c.fetch_pack(p, lambda rs: [], None, None)
    if branch is not None:
        ref = b"refs/headss" + branch.encode("utf-8")
    else:
        ref = b"HEAD"
    try:
        head = result.symrefs[ref]
    except KeyError:
        return None
    if not head.startswith(b"refs/heads/"):
        return head.decode("utf-8")
    return head[len(b"refs/heads/") :].decode("utf-8")


def find_branch(vcs_type, url):
    with VcsWatch() as vcswatch:
        return vcswatch.get_branch_from_url(vcs_type, vcs_git)


with control as updater:
    try:
        vcs_git = updater.source["Vcs-Git"]
    except KeyError:
        pass
    else:
        repo_url, branch, subpath = split_vcs_url(vcs_git)
        try:
            new_branch = find_branch("Git", vcs_git)
        except KeyError:
            pass
        except VcsWatchError as e:
            warn(f"vcswatch URL unusable: {e.args[0]}")
        else:
            if branch != new_branch:
                default_branch = get_default_branch(repo_url)
                if opinionated() or new_branch not in (default_branch, branch):
                    updater.source["Vcs-Git"] = unsplit_vcs_url(
                        repo_url, new_branch, subpath
                    )
                    vcs_browser = determine_browser_url(
                        "Git", updater.source["Vcs-Git"]
                    )
                    if vcs_browser is not None:
                        updater.source["Vcs-Browser"] = vcs_browser
                    report_result("Set branch from vcswatch in Vcs-Git URL.")



#!/usr/bin/python3

from debmutate.deb822 import Deb822Editor

from lintian_brush.fixer import LintianIssue, report_result

try:
    with Deb822Editor(path="debian/copyright") as updater:
        for paragraph in updater.paragraphs:
            if "Files" not in paragraph:
                continue
            if "," not in paragraph["Files"]:
                continue
            if "{" in paragraph["Files"]:
                # Bash-style expansion?
                continue
            issue = LintianIssue(
                "source",
                "comma-separated-files-in-dep5-copyright",
                "paragraph at line XX",
            )
            if issue.should_fix():
                paragraph["Files"] = "\n" + "\n".join(
                    " " + entry.strip()
                    for entry in paragraph["Files"].split(",")
                )
                issue.report_fixed()

except FileNotFoundError:
    pass
except ValueError:
    # Not a machine-readable copyright file
    pass

report_result(
    "debian/copyright: Replace commas with whitespace to separate items "
    "in Files paragraph."
)



#!/usr/bin/python3

import sys

from debmutate.deb822 import Deb822Editor

from lintian_brush.fixer import report_result, vendor
from lintian_brush.lintian import known_tests_control_fields

try:
    from Levenshtein import distance
except ImportError:
    sys.exit(2)


# See https://people.debian.org/~mpitt/autopkgtest/README.package-tests.html
valid_field_names = set(known_tests_control_fields(vendor()))

case_fixed = set()
typo_fixed = set()


try:
    with Deb822Editor("debian/tests/control") as updater:
        for paragraph in updater.paragraphs:
            for field in paragraph:
                if field in valid_field_names:
                    continue
                for option in valid_field_names:
                    if distance(field, option) == 1:
                        value = paragraph[field]
                        del paragraph[field]
                        paragraph[option] = value
                        if option.lower() == field.lower():
                            case_fixed.add((field, option))
                        else:
                            typo_fixed.add((field, option))
                        break
except FileNotFoundError:
    sys.exit(0)

if case_fixed:
    kind = "case" + ("s" if len(case_fixed) > 1 else "")
else:
    kind = ""
if typo_fixed:
    if case_fixed:
        kind += " and "
    kind += "typo" + ("s" if len(typo_fixed) > 1 else "")

fixed_str = ", ".join(
    [
        f"{old} ⇒ {new}"
        for (old, new) in sorted(list(case_fixed) + list(typo_fixed))
    ]
)

report_result(f"Fix field name {kind} in debian/tests/control ({fixed_str}).")



#!/usr/bin/python3

import sys

from debmutate.control import (
    guess_template_type,
)
from debmutate.reformatting import (
    GeneratedFile,
    check_generated_file,
)

from lintian_brush.fixer import LintianIssue, report_result
from lintian_brush.line_editor import LineEditor


def rewrite_line(line):
    if not line.split(b"#", 1)[0].strip():
        return line
    if line.startswith(b" ") or line.startswith(b"\t"):
        return line
    (key, value) = line.split(b":", 1)
    if not value.strip():
        return line
    return b"%s: %s\n" % (key, value.lstrip().rstrip(b"\n"))


def fix_field_spacing(path):
    changed = False
    with LineEditor(path, "b") as e:
        for lineno, oldline in e:
            newline = rewrite_line(oldline)
            if newline != oldline:
                if path == "debian/control":
                    issue = LintianIssue(
                        "source",
                        "debian-control-has-unusual-field-spacing",
                        info=f"line {lineno}"
                    )
                    if issue.should_fix():
                        e[lineno] = newline
                        changed = True
                        issue.report_fixed()
                else:
                    e[lineno] = newline
                    changed = True
    return changed


try:
    check_generated_file("debian/control")
except GeneratedFile as e:
    if e.template_path:
        template_type = guess_template_type(e.template_path, "debian")
        if template_type is None:
            raise
        changed = fix_field_spacing(e.template_path)
        if changed:
            fix_field_spacing("debian/control")
    else:
        raise
except FileNotFoundError:
    sys.exit(0)
else:
    try:
        changed = fix_field_spacing("debian/control")
    except FileNotFoundError:
        sys.exit(0)

if changed:
    report_result("Strip unusual field spacing from debian/control.")



#!/usr/bin/python3

# TODO(jelmer): Read python3 setup.py dist_info
# TODO(jelmer): Check XS-Go-Import-Path

import logging
import os
import sys

from upstream_ontologist import (
    UpstreamMetadata,
    upstream_metadata_sort_key,
)
from upstream_ontologist.guess import (
    check_upstream_metadata,
    extend_upstream_metadata,
    filter_bad_guesses,
    fix_upstream_metadata,
    guess_upstream_metadata_items,
    update_from_guesses,
)
from upstream_ontologist.vcs import convert_cvs_list_to_str

from lintian_brush import (
    min_certainty,
)
from lintian_brush.fixer import (
    LintianIssue,
    current_package_version,
    meets_minimum_certainty,
    net_access_allowed,
    package_is_native,
    report_result,
    trust_package,
)
from lintian_brush.upstream_metadata import (
    ADDON_ONLY_FIELDS,
    DEP12_FIELDS,
)
from lintian_brush.yaml import (
    YamlUpdater,
    update_ordered_dict,
)


def debian_to_upstream_version(v):
    v = v.upstream_version
    if "~" in v:
        return v.split("~")[0]
    if "+" in v:
        return v.split("+")[0]
    return v


# Don't clutter standard err/standard out
logger = logging.getLogger("upstream_ontologist.extrapolate")
logger.setLevel(logging.ERROR)


def filter_by_tag(orig, changed, fields, tag):
    if all(field in orig for field in fields):
        return

    issue = LintianIssue("source", tag, info="[debian/upstream/metadata]")

    if not all(field in orig for field in fields) and not issue.should_fix():
        for field in fields:
            if field in changed:
                del changed[field]

    if all(field in orig or field in changed for field in fields):
        issue.report_fixed()


if package_is_native():
    # Native package
    sys.exit(0)


current_version = current_package_version()

missing_file_issue = LintianIssue(
    "source", "upstream-metadata-file-is-missing", info=()
)

if (
    not os.path.exists("debian/upstream/metadata")
    and not missing_file_issue.should_fix()
):
    sys.exit(0)


with YamlUpdater("debian/upstream/metadata") as editor:
    if isinstance(editor.code, str):
        sys.exit(0)

    code = dict(editor.code)

    repository = code.get("Repository")
    if repository and isinstance(repository, list):
        code["Repository"] = convert_cvs_list_to_str(repository)

    try:
        upstream_metadata = UpstreamMetadata.from_dict(code, "certain")
    except TypeError:
        raise TypeError(f"Invalid upstream metadata: {code!r}")

    minimum_certainty = os.environ.get("MINIMUM_CERTAINTY")
    net_access = net_access_allowed()

    # Downgrade minimum certainty, since check_upstream_metadata can
    # upgrade it to "certain" later.
    initial_minimum_certainty = (
        "likely"
        if net_access and minimum_certainty == "certain"
        else minimum_certainty
    )

    # Do some guessing based on what's in the package
    update_from_guesses(
        upstream_metadata,
        filter_bad_guesses(
            guess_upstream_metadata_items(
                ".",
                trust_package=trust_package(),
                minimum_certainty=initial_minimum_certainty,
            )
        ),
    )

    # Then extend that by contacting e.g. SourceForge
    extend_upstream_metadata(
        upstream_metadata,
        ".",
        minimum_certainty=initial_minimum_certainty,
        net_access=net_access,
        consult_external_directory=True,
    )

    if net_access:
        # Verify that online resources actually exist and adjust certainty
        # accordingly.
        upstream_version = debian_to_upstream_version(current_version)
        check_upstream_metadata(upstream_metadata, version=upstream_version)

    # Homepage is set in debian/control, so don't add it to
    # debian/upstream/metadata.
    external_present_fields = {"Homepage"}

    # If the debian/copyright file is machine-readable, then we do
    # not need to set the Name/Contact information in the
    # debian/upstream/metadata file.
    if "Name" in upstream_metadata or "Contact" in upstream_metadata:
        from debmutate.copyright import upstream_fields_in_copyright

        external_present_fields.update(upstream_fields_in_copyright().keys())

    for key, datum in list(upstream_metadata.items()):
        # Drop keys that don't need to be in debian/upstream/metadata
        if key not in DEP12_FIELDS or key in external_present_fields:
            del upstream_metadata[key]

        # Drop everything that is below our minimum certainty
        elif not meets_minimum_certainty(datum.certainty):
            del upstream_metadata[key]

    achieved_certainty = min_certainty(
        [d.certainty for d in upstream_metadata.values()]
    )

    fix_upstream_metadata(upstream_metadata)

    changed = {
        k: v
        for k, v in upstream_metadata.items()
        if v.value != editor.code.get(k)
    }

    if not changed:
        sys.exit(0)

    filter_by_tag(
        editor.code,
        changed,
        ["Repository", "Repository-Browse"],
        "upstream-metadata-missing-repository",
    )

    filter_by_tag(
        editor.code,
        changed,
        ["Bug-Database", "Bug-Submit"],
        "upstream-metadata-missing-bug-tracking",
    )

    # A change that just says the "Name" field is a bit silly
    if set(changed.keys()) - set(ADDON_ONLY_FIELDS) == {"Name"}:
        sys.exit(0)

    if not os.path.exists("debian/upstream/metadata"):
        missing_file_issue.report_fixed()

    update_ordered_dict(
        editor.code,
        [(k, v.value) for (k, v) in changed.items()],
        key=upstream_metadata_sort_key,
    )

    # If there are only add-on-only fields, then just remove the file.
    if not (set(editor.code.keys()) - set(ADDON_ONLY_FIELDS)):
        editor.code.clear()

    if editor.code and not os.path.isdir("debian/upstream"):
        os.makedirs("debian/upstream", exist_ok=True)


# TODO(jelmer): Add note about other origin fields?
fields = [
    (f"{v.field} (from {v.origin})") if v.origin == "./configure" else v.field
    for k, v in sorted(changed.items())
]

report_result(
    "Set upstream metadata fields: {}.".format(", ".join(sorted(fields))),
    certainty=achieved_certainty,
)



#!/usr/bin/python3

import os

from lintian_brush.fixer import fixed_lintian_tag, report_result

SCRIPTS = ["preinst", "prerm", "postinst", "config", "postrm"]


def replace_set_e(path):
    lines = []
    try:
        with open(path, "rb") as f:
            lines = list(f.readlines())
    except (FileNotFoundError, IsADirectoryError):
        return

    if b"set -e\n" in lines:
        return

    # TODO(jelmer): Handle -e in combination with other flags.
    if lines[0] != b"#!/bin/sh -e\n":
        return
    lines[0] = b"#!/bin/sh\n"
    for i, line in enumerate(lines[1:]):
        if line == b"set -e\n":
            return
        if (
            not (line.startswith(b"#") or line == b"\n")
            or line.strip() == b"#DEBHELPER#"
        ):
            fixed_lintian_tag(
                "source",
                "maintainer-script-without-set-e",
                os.path.basename(path),
            )
            if not lines[i - 1].strip():
                lines.insert(i, b"set -e\n")
                lines.insert(i + 1, b"\n")
            else:
                lines.insert(i, b"\n")
                lines.insert(i + 1, b"set -e\n")
            break

    with open(path, "wb") as f:
        f.writelines(lines)


for name in SCRIPTS:
    replace_set_e(os.path.join("debian", name))


report_result("Use set -e rather than passing -e on the shebang-line.")



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Utility functions for dealing with Vcs URLs of various types."""

__all__ = [
    "fixup_broken_git_url",
    "sanitize_url",
    "determine_browser_url",
]


from typing import Callable, List, Optional, Union

from debmutate.vcs import split_vcs_url, unsplit_vcs_url
from upstream_ontologist.vcs import (
    canonical_git_repo_url,
    convert_cvs_list_to_str,
    drop_vcs_in_scheme,
    find_public_repo_url,
    find_secure_repo_url,
    fixup_broken_git_details,
    fixup_rcp_style_git_repo_url,
)

from . import _lintian_brush_rs


def find_public_vcs_url(url: str) -> Optional[str]:
    (repo_url, branch, subpath) = split_vcs_url(url)
    revised_url = find_public_repo_url(repo_url)
    if revised_url is not None:
        return unsplit_vcs_url(revised_url, branch, subpath)
    return None


def fixup_rcp_style_git_url(url: str) -> str:
    (repo_url, branch, subpath) = split_vcs_url(url)
    repo_url = fixup_rcp_style_git_repo_url(repo_url)
    return unsplit_vcs_url(repo_url, branch, subpath)


determine_gitlab_browser_url = _lintian_brush_rs.determine_gitlab_browser_url
determine_browser_url = _lintian_brush_rs.determine_browser_url
canonicalize_vcs_browser_url = _lintian_brush_rs.canonicalize_vcs_browser_url


def canonical_vcs_git_url(url: str) -> str:
    (repo_url, branch, subpath) = split_vcs_url(url)
    repo_url = canonical_git_repo_url(repo_url)
    return unsplit_vcs_url(repo_url, branch, subpath)


canonicalize_vcs_fns = {
    "Browser": canonicalize_vcs_browser_url,
    "Git": canonical_vcs_git_url,
}


def canonicalize_vcs_url(vcs_type: str, url: str) -> str:
    try:
        return canonicalize_vcs_fns[vcs_type](url)
    except KeyError:
        return url


def find_secure_vcs_url(url: str, net_access: bool = True) -> Optional[str]:
    repo_url: Optional[str]
    (repo_url, branch, subpath) = split_vcs_url(url)
    repo_url = find_secure_repo_url(
        repo_url, branch=branch, net_access=net_access
    )
    if repo_url is None:
        return None

    return unsplit_vcs_url(repo_url, branch, subpath)


def fixup_broken_git_url(url: str) -> str:
    """Attempt to fix up broken Git URLs.

    A common misspelling is to add an extra ":" after the hostname
    """
    repo_url, branch, subpath = split_vcs_url(url)
    newrepo_url, newbranch, newsubpath = fixup_broken_git_details(
        repo_url, branch, subpath
    )
    if newrepo_url != repo_url or newbranch != branch or newsubpath != subpath:
        return unsplit_vcs_url(newrepo_url, newbranch, newsubpath)
    return url


SANITIZERS: List[Callable[[str], str]] = [
    drop_vcs_in_scheme,
    fixup_broken_git_url,
    fixup_rcp_style_git_url,
    lambda url: find_public_vcs_url(url) or url,
    canonical_vcs_git_url,
    lambda url: find_secure_vcs_url(url, net_access=False) or url,
]


def sanitize_url(url: Union[str, List[str]]) -> str:
    """Sanitize a version control URL."""
    if isinstance(url, list):
        url_str = convert_cvs_list_to_str(url)
    else:
        url_str = url
    url_str = url_str.strip()
    for sanitizer in SANITIZERS:
        url_str = sanitizer(url_str)
    return url_str



#!/usr/bin/python3
import os
import sys

from lintian_brush.fixer import LintianIssue, report_result

if not os.path.exists("debian/pyversions"):
    sys.exit(2)

with open("debian/pyversions") as f:
    pyversions = f.read().strip()

if pyversions.startswith("2."):
    issue = LintianIssue("source", "debian-pyversions-is-obsolete")
    if issue.should_fix():
        os.unlink("debian/pyversions")
        issue.report_fixed()

report_result("Remove obsolete debian/pyversions.")



#!/usr/bin/python3

from contextlib import suppress

from debmutate.reformatting import check_generated_file

from lintian_brush.fixer import LintianIssue, report_result
from lintian_brush.line_editor import LineEditor

check_generated_file("debian/rules")

with suppress(FileNotFoundError), LineEditor("debian/rules", "b") as e:
    for lineno, line in e:
        newline = line.replace(b"$(PWD)", b"$(CURDIR)")
        if newline != line:
            issue = LintianIssue(
                "source", "debian-rules-calls-pwd", info=f"line {lineno}"
            )
            if issue.should_fix():
                e[lineno] = newline
                issue.report_fixed()

report_result("debian/rules: Avoid using $(PWD) variable.")



#!/usr/bin/python3

import os
import sys

from lintian_brush.fixer import report_result

if not os.path.exists("debian/rules"):
    sys.exit(0)

os.chmod("debian/rules", 0o755)

report_result("Mark debian/rules as executable.", certainty="certain")



#!/usr/bin/python3
# Copyright (C) 2018 Jelmer Vernooij <jelmer@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# See https://wiki.debian.org/UpstreamMetadata and
# https://dep-team.pages.debian.net/deps/dep12/

ADDON_ONLY_FIELDS = ["Archive"]
DEP12_FIELDS = [
    "Name",
    "Contact",
    "Archive",
    "ASCL-Id",
    "Bug-Database",
    "Bug-Submit",
    "Cite-As",
    "Changelog",
    "CPE",
    "Documentation",
    "Donation",
    "FAQ",
    "Funding",
    "Gallery",
    "Other-References",
    "Reference",
    "Registration",
    "Registry",
    "Repository",
    "Repository-Browse",
    "Screenshots",
    "Security-Contact",
    "Webservice",
]



#!/usr/bin/python3

import sys

try:
    import psycopg2  # noqa: F401
except ModuleNotFoundError:
    sys.exit(2)

from lintian_brush.fixer import (
    control,
    diligence,
    fixed_lintian_tag,
    net_access_allowed,
    report_result,
)
from lintian_brush.lintian_overrides import remove_unused

INTERMITTENT_LINTIAN_TAGS = [
    "rc-version-greater-than-expected-version",
]


if diligence() < 1:
    # Removing unused overrides requires pro-actively contacting UDD.
    sys.exit(0)

if not net_access_allowed():
    sys.exit(0)

removed = remove_unused(
    control.paragraphs, ignore_tags=INTERMITTENT_LINTIAN_TAGS
)

description = [
    f"Remove {len(removed)} unused lintian overrides.\n"
    "\n",
]
for override in removed:
    description.append(f"* {override.tag}\n")
    fixed_lintian_tag(
        "source",
        "unused-override",
        info=(override.tag, override.info if override.info else ""),
    )

report_result("".join(description), certainty="certain")



#!/usr/bin/python3

import re
import socket
from functools import partial

from debmutate.changelog import ChangelogEditor

from lintian_brush import min_certainty
from lintian_brush.fixer import (
    fixed_lintian_tag,
    fixed_lintian_tags,
    meets_minimum_certainty,
    net_access_allowed,
    report_result,
    warn,
)

certainty = "certain"
debbugs = None


def valid_bug(package, bug):
    if not net_access_allowed():
        return None
    global debbugs
    if debbugs is None:
        from lintian_brush.debbugs import DebBugs

        _debbugs = DebBugs()
        try:
            _debbugs.connect()
        except ImportError:
            # No asynpcg?
            return None
        except socket.gaierror as e:
            warn(f"Unable to connect to debbugs: {e}")
            return None
        debbugs = _debbugs
    return debbugs.check_bug(package, bug)


def check_bug(package, bugno):
    valid = valid_bug(package, bugno)
    if valid is not None:
        return (valid, "certain")
    # Let's assume valid, but downgrade certainty
    valid = True
    # Check number of digits; upstream projects don't often hit the 5-digit
    # bug numbers that Debian has.
    if len(str(bugno)) >= 5:
        certainty = "likely"
    else:
        certainty = "possible"
    return (valid, certainty)


def fix_close_colon(package, m):
    global certainty
    bugno = int(m.group("bug"))
    (valid, bug_certainty) = check_bug(package, bugno)
    if meets_minimum_certainty(bug_certainty) and valid:
        certainty = min_certainty([certainty, bug_certainty])
        fixed_lintian_tag(
            "all",
            "possible-missing-colon-in-closes",
            info=f"{m.group('closes')} #{bugno}"
        )
        return f"{m.group('closes')}: #{bugno}"
    else:
        return m.group(0)


def fix_close_typo(package, m):
    global certainty
    bugno = int(m.group("bug"))
    (valid, bug_certainty) = check_bug(package, bugno)
    if meets_minimum_certainty(bug_certainty) and valid:
        certainty = min_certainty([certainty, bug_certainty])
        fixed_lintian_tag(
            "all", "misspelled-closes-bug", info="#{}".format(m.group("bug"))
        )
        return "{}s: #{}".format(m.group("close"), m.group("bug"))
    else:
        return m.group(0)


with ChangelogEditor() as updater:
    for block in updater.changelog:
        for i, change in enumerate(block._changes):
            change = re.sub(
                r"(?<!partially )(?P<closes>closes) " r"#(?P<bug>[0-9]+)",
                partial(fix_close_colon, block.package),
                change,
                flags=re.IGNORECASE,
            )
            change = re.sub(
                "(?P<close>close): #(?P<bug>[0-9]+)",
                partial(fix_close_typo, block.package),
                change,
                flags=re.IGNORECASE,
            )
            block._changes[i] = change


if fixed_lintian_tags() == {"possible-missing-colon-in-closes"}:
    report_result("Add missing colon in closes line.", certainty=certainty)
elif fixed_lintian_tags() == {"misspelled-closes-bug"}:
    report_result("Fix misspelling of Close ⇒ Closes.", certainty=certainty)
else:
    report_result("Fix formatting of bug closes.", certainty=certainty)



#!/usr/bin/python3

from lintian_brush.fixer import control, fixed_lintian_tag, report_result
from lintian_brush.vcs import determine_browser_url

with control as updater:
    if "Vcs-Browser" not in updater.source:
        try:
            vcs_git = updater.source["Vcs-Git"]
        except KeyError:
            pass
        else:
            browser_url = determine_browser_url("git", vcs_git)
            if browser_url is not None:
                updater.source["Vcs-Browser"] = browser_url
                fixed_lintian_tag(
                    updater.source,
                    "missing-vcs-browser-field",
                    info=f"Vcs-Git {vcs_git}",
                )


report_result("debian/control: Add Vcs-Browser field")



#!/usr/bin/python
# Copyright (C) 2018 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Automatically fix lintian issues."""

import logging
import os
import sys
from typing import (
    Optional,
    Sequence,
)

import breezy.bzr  # noqa: F401
import breezy.git  # noqa: F401
from breezy.workingtree import WorkingTree

__version__ = (0, 160)
version_string = ".".join(map(str, __version__))
SUPPORTED_CERTAINTIES = ["certain", "confident", "likely", "possible", None]
DEFAULT_MINIMUM_CERTAINTY = "certain"
USER_AGENT = "lintian-brush/" + version_string
# Too aggressive?
DEFAULT_URLLIB_TIMEOUT = 3
logger = logging.getLogger(__name__)


class UnsupportedCertainty(Exception):
    """Unsupported certainty."""


def min_certainty(certainties: Sequence[str]) -> str:
    return confidence_to_certainty(
        max([SUPPORTED_CERTAINTIES.index(c) for c in certainties] + [0])
    )


def confidence_to_certainty(confidence: Optional[int]) -> str:
    if confidence is None:
        return "unknown"
    try:
        return SUPPORTED_CERTAINTIES[confidence] or "unknown"
    except IndexError as exc:
        raise ValueError(confidence) from exc


def certainty_sufficient(
    actual_certainty: str, minimum_certainty: Optional[str]
) -> bool:
    """Check if the actual certainty is sufficient.

    Args:
      actual_certainty: Actual certainty with which changes were made
      minimum_certainty: Minimum certainty to keep changes
    Returns:
      boolean
    """
    actual_confidence = certainty_to_confidence(actual_certainty)
    if actual_confidence is None:
        # Actual confidence is unknown.
        # TODO(jelmer): Should we really be ignoring this?
        return True
    minimum_confidence = certainty_to_confidence(minimum_certainty)
    if minimum_confidence is None:
        return True
    return actual_confidence <= minimum_confidence


def open_binary(name):
    return open(data_file_path(name), "rb")  # noqa: SIM115


def data_file_path(name, check=os.path.exists):
    # There's probably a more Pythonic way of doing this, but
    # I haven't bothered finding out what it is yet..
    path = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "..", name)
    )
    if check(path):
        return path

    path = os.path.abspath(
        os.path.join(
            os.path.dirname(__file__), "..", "..", "lintian-brush", name
        )
    )
    if check(path):
        return path

    path = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "..", "analyzer", name)
    )
    if check(path):
        return path

    import pkg_resources

    path = pkg_resources.resource_filename(__name__, f"lintian-brush/{name}")
    if check(path):
        return path

    # Urgh.
    for b in [
        "/usr/share/lintian-brush",
        "/usr/local/share/lintian-brush",
        os.path.join(sys.prefix, "share/lintian-brush"),
    ]:
        path = os.path.join(b, name)
        if check(path):
            return path
    raise RuntimeError(f"unable to find data path: {name}")


def get_committer(tree: WorkingTree) -> str:
    """Get the committer string for a tree.

    Args:
      tree: A Tree object
    Returns:
      A committer string
    """
    # TODO(jelmer): Perhaps this logic should be in Breezy?
    if hasattr(tree.branch.repository, "_git"):
        cs = tree.branch.repository._git.get_config_stack()
        user = os.environ.get("GIT_COMMITTER_NAME")
        email = os.environ.get("GIT_COMMITTER_EMAIL")
        if user is None:
            try:
                user = cs.get(("user",), "name").decode("utf-8")
            except KeyError:
                user = None
        if email is None:
            try:
                email = cs.get(("user",), "email").decode("utf-8")
            except KeyError:
                email = None
        if user and email:
            return user + " <" + email + ">"
        from breezy.config import GlobalStack

        return GlobalStack().get("email")
    else:
        config = tree.branch.get_config_stack()
        return config.get("email")


def certainty_to_confidence(certainty: Optional[str]) -> Optional[int]:
    if certainty in ("unknown", None):
        return None
    return SUPPORTED_CERTAINTIES.index(certainty)


def fixable_lintian_tags():
    from ruamel.yaml import YAML

    yaml = YAML()
    path = os.path.join(
        os.path.dirname(__file__), "../../lintian-brush/fixers/index.desc"
    )
    with open(path) as f:
        fixers = yaml.load(f)

    supported_tags = set()
    for fixer in fixers:
        try:
            tags = fixer["lintian-tags"]
        except KeyError:
            pass
        else:
            if tags is not None:
                supported_tags.update(tags)

    return supported_tags



#!/usr/bin/python3

from contextlib import suppress

from debmutate.copyright import CopyrightEditor, NotMachineReadableError

from lintian_brush.fixer import LintianIssue, report_result

with suppress(
    FileNotFoundError, NotMachineReadableError
), CopyrightEditor() as editor:
    files_i = 0
    for i, paragraph in enumerate(editor.copyright.all_files_paragraphs()):
        if "Files" in paragraph:
            if paragraph["Files"] == "*" and files_i > 0:
                issue = LintianIssue(
                    "source",
                    "global-files-wildcard-not-first-paragraph-in-"
                    "dep5-copyright",
                )
                if issue.should_fix():
                    editor.insert(0, editor.pop(i))
                    issue.report_fixed()
            files_i += 1

report_result('Make "Files: *" paragraph the first in the copyright file.')



#!/usr/bin/python3

import sys
from contextlib import suppress

from debmutate.control import ensure_some_version, get_relation

from lintian_brush.fixer import LintianIssue, control, report_result

try:
    from tomlkit import load
except ModuleNotFoundError:
    sys.exit(2)

try:
    with open("pyproject.toml") as f:
        toml = load(f)
except FileNotFoundError:
    sys.exit(0)


build_backend = toml.get("build-system", {}).get("build-backend")

# See /usr/share/lintian/lib/Lintian/Check/Languages/Python.pm
PREREQUISITE_MAP = {
    "poetry.core.masonry.api": "python3-poetry-core",
    "flit_core.buildapi": "flit",
    "setuptools.build_meta": "python3-setuptools",
}


try:
    prerequisite = PREREQUISITE_MAP[build_backend]
except KeyError:
    sys.exit(2)

with control:
    for field in [
        "Build-Depends",
        "Build-Depends-Indep",
        "Build-Depends-Arch",
    ]:
        with suppress(KeyError):
            if get_relation(control.source.get(field, ""), prerequisite):
                sys.exit(0)
    # TOOD(jelmer): Add file:lineno; requires
    # https://github.com/sdispater/tomlkit/issues/55
    issue = LintianIssue(
        control.source,
        "missing-prerequisite-for-pyproject-backend",
        info=f"{build_backend} (does not satisfy {prerequisite})",
    )
    if issue.should_fix():
        control.source["Build-Depends"] = ensure_some_version(
            control.source.get("Build-Depends", ""), prerequisite
        )
        issue.report_fixed()

report_result(
    f"Add missing build-dependency on {prerequisite}.\n\n"
    f"This is necessary for build-backend {build_backend} in pyproject.toml"
)



#!/usr/bin/python3

import sys

from debmutate.control import delete_from_list

from lintian_brush.fixer import control, fixed_lintian_tag, report_result

# TODO(jelmer): support checking debcargo's maintainer/uploaders fields

try:
    with control as updater:
        if "Uploaders" in updater.source:
            uploaders = updater.source["Uploaders"].split(",")
            maintainer = updater.source["Maintainer"]
            if maintainer in [uploader.strip() for uploader in uploaders]:
                updater.source["Uploaders"] = delete_from_list(
                    updater.source["Uploaders"], maintainer
                )
                if not updater.source["Uploaders"].strip():
                    del updater.source["Uploaders"]
                fixed_lintian_tag(
                    updater.source, "maintainer-also-in-uploaders"
                )
except FileNotFoundError:
    sys.exit(0)


report_result("Remove maintainer from uploaders.")



#!/usr/bin/python3

import re

from lintian_brush.fixer import LintianIssue, report_result

with open("debian/copyright", "rb") as f:
    lines = list(f)

m = re.match(
    rb"^(Format|Format-Specification): "
    rb"(http:\/\/www.debian.org\/doc\/packaging-manuals\/"
    rb"copyright-format\/1.0.*)\n",
    lines[0],
)
if m:
    newline = (
        b"Format: https://www.debian.org/doc/packaging-manuals/"
        b"copyright-format/1.0/\n"
    )
    if newline != lines[0]:
        lines[0] = newline
        issue = LintianIssue(
            "source", "insecure-copyright-format-uri", m.group(2).decode()
        )
        if issue.should_fix():
            with open("debian/copyright", "wb") as f:
                f.writelines(lines)
            issue.report_fixed()

report_result("Use secure copyright file specification URI.")
